#include <signal.h>
#include <be/app/Message.h>
#include <be/interface/Button.h>
#include <be/interface/ListView.h>
#include <be/interface/Rect.h>
#include <be/interface/StatusBar.h>
#include <be/interface/StringView.h>
#include <be/storage/Directory.h>
#include <be/storage/Entry.h>
#include <be/storage/Node.h>
#include <be/storage/NodeInfo.h>
#include <be/support/Debug.h>
#include <be/support/List.h>
#include <be/support/String.h>

#include <be/support/Beep.h> // for test purposes, by CP, 20030414

#include <EntryRefItem/EntryRefItem.h>
#include <PickListView/PickListView.h>
#include <Santa/BetterScrollView.h>
#include "AddOnView.h"
#include "AppDefs.h"
#include "AppView.h"
#include "CommandConstants.h"
#include "EditorView.h"
#include "GUIStrings.h"
#include "NAView.h"
#include "Preferences.h"
#include "TAView.h"

AppView::AppView(BRect frame) : BView(frame, "AppView", B_FOLLOW_ALL,
		B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE_JUMP)
{
	PRINT(("AppView::AppView(BRect)\n"));

	applying = false;
}

void
AppView::InitView()
{
	PRINT(("AppView::InitView()\n"));

	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));

	float space = 6;

	BRect frame(0,0,100,100);
	pickListView = new PickListView(frame,"pickListView",
			B_FOLLOW_LEFT | B_FOLLOW_TOP);
	AddChild(pickListView);

	BRect inframe = pickListView->BoxBounds();
	pickListView->AddView(new EditorView(inframe));
	pickListView->AddView(new TAView(inframe));
	pickListView->AddView(new NAView(inframe));

	if(prefs->Lock())
	{
		SelectView(prefs->GetPrefFunction());
		prefs->Unlock();
	}

	frame = pickListView->Frame();
	frame.bottom = Bounds().bottom - 50;
	frame.OffsetBy(frame.Width() + space,0);
	frame.InsetBy(0,12);
	frame.right = Bounds().right - space - B_V_SCROLL_BAR_WIDTH;
	frame.bottom -=  B_H_SCROLL_BAR_HEIGHT;
	listView = new BListView(frame,"listView",B_MULTIPLE_SELECTION_LIST,
			B_FOLLOW_ALL);
	listView->SetSelectionMessage(new BMessage(SELECTION_CHANGED));

	scrollView = new BetterScrollView("scrollView",listView,B_FOLLOW_ALL);
	scrollView->SetDataRect(BRect(0,0,0,0));
	AddChild(scrollView);

	BRect textFrame = scrollView->Frame();
	textFrame.bottom += 15;
	textFrame.top = textFrame.bottom - 10;
	textFrame.left = textFrame.right - 150;
	selectedStringView = new BStringView(textFrame,"selectedStringView","",
			B_FOLLOW_RIGHT|B_FOLLOW_BOTTOM);
	selectedStringView->SetAlignment(B_ALIGN_RIGHT);
	selectedStringView->SetFontSize(10);
	AddChild(selectedStringView);

	frame = Bounds();
	frame.InsetBy(space,space);
	frame.top = frame.bottom - 30;
	frame.left = frame.right - 70;
	applyButton = new BButton(frame,"applyButton",APPLY_BTN,new BMessage(APPLY),
			B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM);
	applyButton->SetEnabled(false);
	AddChild(applyButton);

	frame.OffsetBy(-(frame.Width() + space),0);
	resetButton = new BButton(frame,"resetButton",RESET_BTN,new BMessage(RESET),
			B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM);
	resetButton->SetEnabled(false);
	AddChild(resetButton);

	frame = Bounds();
	frame.InsetBy(space,space);
	frame.right = (resetButton->Frame()).left - space;
	frame.top = frame.bottom - 35;
	statusBar = new BStatusBar(frame,"statusBar","0%","100%");
	statusBar->SetResizingMode(B_FOLLOW_LEFT_RIGHT|B_FOLLOW_BOTTOM);
	//statusBar->Hide();
	AddChild(statusBar);
}

// This function added by CP, 20030414
/* Select next item in list, provided that:
**   - only one item is selected
**   - the selected item is not the last item 
** If no item is currently selected, select the first item in list.
*/
void AppView::NextItem()
{
	int32 selectedItem;
	// if the list of selected items has more than one entry, there
	// are multiple selected items; we just return in this case
	if( listView->CurrentSelection(1) >= 0 )
	{
		beep();
		return;
	}
	// fetch selected entry
	selectedItem = listView->CurrentSelection(0);
	// if no item is selected, select first
	if( selectedItem < 0 )
	{
		listView->Select(0);
		return;
	}
	// check if selected item is last item
	if( selectedItem+1 == listView->CountItems() )
	{
		beep();
		return;
	}
	// Okayovich, select next item. Select will automatically deselect previous
	// selections if called without extend flag == true
	listView->Select(selectedItem+1);
}

// This function added by CP, 20030414
/* Select previous item in list, provided that:
**   - only one item is selected
**   - the selected item is not the first item 
** If no item is currently selected, select the last item in list.
*/
void AppView::PrevItem()
{
	int32 selectedItem;
	// if the list of selected items has more than one entry, there
	// are multiple selected items; we just return in this case
	if( listView->CurrentSelection(1) >= 0 )
	{
		beep();
		return;
	}
	// fetch selected entry
	selectedItem = listView->CurrentSelection(0);
	// if no item is selected, select last
	if( selectedItem < 0 )
	{
		listView->Select(listView->CountItems() - 1);
		return;
	}
	// check if selected item is first item
	if( selectedItem == 0 )
	{
		beep();
		return;
	}
	// Select previous item. Select will automatically deselect previous
	// selections if called without extend flag == true
	listView->Select(selectedItem-1);
}
// This function added by CP, 20031015
void AppView::FirstItem() {
	listView->Select(0);
}
// This function added by CP, 20031015
void AppView::LastItem() {
	int32 count = listView->CountItems();
	if(count > 0)
		listView->Select(count-1);
}

/* Select all list entries. This is accessed by menu item from AppWindow. */
void
AppView::SelectAllItems()
{
	listView->Select(0, listView->CountItems()-1);
}

void
AppView::ClearList()
{
	PRINT(("AppView::ClearList()\n"));

	if(!applying)
	{
		listView->DeselectAll();
		int numItems = listView->CountItems();
		for(int i=0;i<numItems;i++)
		{
			BListItem* item = listView->RemoveItem((int32)0);
			delete item;
		}
		listView->Invalidate();
		scrollView->SetDataRect(BRect(0,0,0,0));
	}
}

void
AppView::SelectView(int32 index)
{
	PRINT(("AppView::SelectView(int32)\n"));

	pickListView->SelectView(index);
}

void
AppView::SelectionChanged()
{
	PRINT(("AppView::SelectionChanged()\n"));

	if(!applying)
	{
		int numViews = pickListView->CountViews();

		if(numViews > 0)
		{
			BList list;
			int i=0;
			int selected;
			while((selected = listView->CurrentSelection(i)) >= 0)
			{
				list.AddItem(listView->ItemAt(selected));
				i++;
			}

			BString buf;
			buf << list.CountItems();
			buf << " / ";
			buf << listView->CountItems();
			buf << SELECTED_TEXT;
			selectedStringView->SetText(buf.String());

			if(list.CountItems() > 0)
			{
				applyButton->SetEnabled(true);
				resetButton->SetEnabled(true);
			}
			else
			{
				applyButton->SetEnabled(false);
				resetButton->SetEnabled(false);
			}

			for(int i=0;i<numViews;i++)
			{
				AddOnView* view = (AddOnView*)pickListView->ViewAt(i);
				if(view)
				{
					view->SelectionChanged(&list);
				}
			}
		}
	}
}

void
AppView::Apply()
{
	PRINT(("AppView::Apply()\n"));

	AddOnView* view = (AddOnView*)pickListView->SelectedView();
	if(view)
	{
		view->Apply();
	}
}

void
AppView::Reset()
{
	PRINT(("AppView::Reset()\n"));

	// Select All
	int numItems = listView->CountItems();
	listView->Select(0,(int32)numItems-1);

	AddOnView* view = (AddOnView*)pickListView->SelectedView();
	if(view)
	{
		view->Reset();
	}
}

void
AppView::AddRefs(BMessage* refsMessage)
{
	PRINT(("AppView::AddRefs(BMessage*)\n"));

	BList* args = new BList();
	BMessage* msg = new BMessage(*refsMessage);
	args->AddItem(this);
	args->AddItem(msg);

	thread_id my_thread = spawn_thread(&AppView::AddRefsThreadFunc,"AddRefs",
			B_NORMAL_PRIORITY,(void*)args);
	resume_thread(my_thread);
}

int32
AppView::AddRefsThreadFunc(void* data)
{
	PRINT(("AppView::AddRefsThreadFunc(void*)\n"));

	signal(SIGQUIT,quit_requested);

	BList* dataList = (BList*)data;
	AppView* view = (AppView*)dataList->ItemAt(0);
	BMessage* refsMessage = (BMessage*)dataList->ItemAt(1);

	if(view->LockLooper())
	{
		view->listView->DeselectAll();
		view->UnlockLooper();
	}
	
	view->AddRefsHelper(view,refsMessage);

	if(view->LockLooper())
	{
		// Setup DataRect for scrollView
		int numItems = view->listView->CountItems();
		float frameWidth = 0.0;
		float frameHeight= 0.0;
		for(int i=0;i<numItems;i++)
		{
			BRect frame = view->listView->ItemFrame(i);
			frameHeight += (frame.Height() + 1.00);

			EntryRefItem* item = (EntryRefItem*)view->listView->ItemAt(i);
			float width = view->listView->StringWidth(item->Text());
			if(width > frameWidth)
			{
				frameWidth = width;
			}
		}
		BRect dataRect(0,0,frameWidth,frameHeight);
		view->scrollView->SetDataRect(dataRect);

		view->listView->SortItems(&AppView::SortFunc);

		// Select All
		view->listView->Select(0,(int32)numItems-1);

		view->UnlockLooper();
	}

	return(B_OK);
}

void
AppView::AddRefsHelper(AppView* view, BMessage* refsMessage)
{
	PRINT(("AppView::AddRefsHelper(AppView*,BMessage*)\n"));

	entry_ref tmp_ref;
	int numRefs = 0;
	while(refsMessage->FindRef("refs",numRefs,&tmp_ref) == B_NO_ERROR)
	{
		BNode node(&tmp_ref);
		if(node.IsFile())
		{
			BNodeInfo info(&node);
			char type[B_MIME_TYPE_LENGTH+1];
			info.GetType(type);
			if(strncmp(type,"audio",5) == 0)
			{
				bool found = false;
				for(int i=0;i<listView->CountItems();i++)
				{
					EntryRefItem* item = (EntryRefItem*)listView->ItemAt(i);
					entry_ref* ref = item->EntryRef();
					if(*ref == tmp_ref)
					{
						found = true;
						break;
					}
				}
				if(!found)
				{
					if(view->LockLooper())
					{
						view->listView->AddItem(new EntryRefItem(&tmp_ref));
						view->UnlockLooper();
					}
				}
			}
		}
		else if(node.IsDirectory())
		{
			BDirectory dir(&tmp_ref);
			if(dir.InitCheck() == B_OK)
			{
				BMessage message(B_REFS_RECEIVED);
				entry_ref ref;
				while(dir.GetNextRef(&ref) != B_ENTRY_NOT_FOUND)
				{
					message.AddRef("refs",&ref);
				}
				AddRefsHelper(view,&message);
			}
		}
		else if(node.IsSymLink())
		{
			BEntry entry(&tmp_ref,true);
			if((entry.InitCheck() == B_OK) && (entry.GetRef(&tmp_ref) == B_OK))
			{
				BMessage message(B_REFS_RECEIVED);
				message.AddRef("refs",&tmp_ref);
				AddRefsHelper(view,&message);
			}
		}
		numRefs++;
	}
}

int
AppView::SortFunc(const void* first, const void* second)
{
	//PRINT(("AppWindow::SortFunc(const void*,const void*)\n"));
	EntryRefItem* firstItem = *((EntryRefItem**)first);
	EntryRefItem* secondItem = *((EntryRefItem**)second);
	return strcmp(firstItem->Text(),secondItem->Text());
}

void
AppView::AttachedToWindow()
{
	PRINT(("AppView::AttachedToWindow()\n"));

	BView::AttachedToWindow();

	InitView();

	listView->SetTarget(this);
	applyButton->SetTarget(this);
	resetButton->SetTarget(this);
}

void
AppView::GetPreferredSize(float* width, float* height)
{
	int space = 10;

	*width = Bounds().Width();
	*height = pickListView->Frame().Height() + statusBar->Frame().Height() +
		3*space;
}

void
AppView::MessageReceived(BMessage* message)
{
	//PRINT(("AppView::MessageReceived(BMessage*)\n"));
	switch(message->what)
	{
		case SELECTION_CHANGED:
			SelectionChanged();
			break;
		case APPLY:
			Apply();
			break;
		case RESET:
			Reset();
			break;
		case START_APPLY:
			applying = true;
			statusBar->Show();
			applyButton->SetEnabled(false);
			resetButton->SetEnabled(false);
			break;
		case END_APPLY:
			statusBar->Hide();
			applyButton->SetEnabled(true);
			resetButton->SetEnabled(true);
			applying = false;
			SelectionChanged();
			break;
		case STATUS_BAR_SET_MAX_VALUE:
			int32 maxValue;
			message->FindInt32("maxValue",&maxValue);
			statusBar->SetMaxValue(maxValue);
			break;
		case STATUS_BAR_UPDATE:
			statusBar->Update(1);
			break;
		case STATUS_BAR_RESET:
			statusBar->Reset("0%","100%");
			break;
		case INVALIDATE_LIST:
			listView->Invalidate();
			listView->SortItems(&AppView::SortFunc);
			break;
		case B_REFS_RECEIVED:
			AddRefs(message);
			break;
		case B_SIMPLE_DATA:
			AddRefs(message);
			break;
		case CHANGE_VIEW:
			{
				PRINT(("got here view\n"));
				int32 index;
				message->FindInt32("index",&index);
				SelectView(index);
			}
			break;
		default:
			BView::MessageReceived(message);
	}
}

void quit_requested(int sig)
{
	exit_thread(0);
}

